# django imports
from django import template
from django.contrib.contenttypes.models import ContentType

# review imports
from reviews import utils as reviews_utils

register = template.Library()


@register.inclusion_tag('reviews/reviews_for_instance.html', takes_context=True)
def reviews_for_instance(context, instance):
    """
    """
    request = context.get("request")
    ctype = ContentType.objects.get_for_model(instance)
    has_rated = reviews_utils.has_rated(request, instance)
    reviews = reviews_utils.get_reviews_for_instance(instance)

    return {
        "reviews": reviews,
        "has_rated": has_rated,
        "content_id": instance.id,
        "content_type_id": ctype.id,
        "MEDIA_URL": context.get("MEDIA_URL")
    }


@register.inclusion_tag('reviews/average_for_instance.html', takes_context=True)
def average_for_instance(context, instance):
    """
    """
    average, amount = reviews_utils.get_average_for_instance(instance)
    return {
        "average": average,
        "amount": amount,
    }



# django imports
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist
from django.db import connection

# review imports
from reviews.models import Review


def get_best_rated():
    """Returns the best rated instance for all models.
    """
    cursor = connection.cursor()
    cursor.execute("""SELECT avg(score), content_type_id, content_id
                      FROM reviews_review
                      WHERE active=%s
                      GROUP BY content_id
                      ORDER BY score DESC""", [True])

    try:
        score, content_type_id, content_id = cursor.fetchone()
        ctype = ContentType.objects.get_for_id(content_type_id)
        content = ctype.model_class().objects.get(pk=content_id)
        return content, score
    except (TypeError, ObjectDoesNotExist):
        return None


def get_best_rated_for_model(instance):
    """Returns the best rated instance for given model or instance of a model.
    """
    ctype = ContentType.objects.get_for_model(instance)

    cursor = connection.cursor()
    cursor.execute("""SELECT avg(score), content_id
                      FROM reviews_review
                      WHERE content_type_id=%s
                      AND active=%s
                      GROUP BY content_id
                      ORDER BY score DESC""", [ctype.id, True])

    try:
        score, content_id = cursor.fetchone()
        content = ctype.model_class().objects.get(pk=content_id)
        return content, score
    except (TypeError, ObjectDoesNotExist):
        return None


def get_reviews_for_instance(instance):
    """Returns active reviews for given instance.
    """
    ctype = ContentType.objects.get_for_model(instance)
    return Review.objects.active().filter(content_type=ctype.id, content_id=instance.id)


def get_average_for_instance(instance):
    """Returns the average score and the amount of reviews for the given
    instance. Takes only active reviews into account.

    Returns (average, amount)
    """
    # TODO: Check Django 1.1's aggregation
    ctype = ContentType.objects.get_for_model(instance)
    cursor = connection.cursor()
    cursor.execute("""SELECT avg(score), count(*)
                      FROM reviews_review
                      WHERE content_type_id=%s
                      AND content_id=%s
                      AND active=%s""", [ctype.id, instance.id, True])

    return cursor.fetchone()


def has_rated(request, instance):
    """Returns True if the current user has already rated for the given
    instance.
    """
    ctype = ContentType.objects.get_for_model(instance)

    try:
        if request.user.is_authenticated():
            Review.objects.get(
                content_type=ctype.id,
                content_id=instance.id,
                user=request.user
            )
        else:
            Review.objects.get(
                content_type=ctype.id,
                content_id=instance.id,
                session_id=request.session.session_key
            )
    except ObjectDoesNotExist:
        return False
    else:
        return True



from django.conf.urls import patterns, url

# Catalog
urlpatterns = patterns('reviews.views',
    url(r'^add/(?P<content_type_id>\d*)/(?P<content_id>\d*)$', "add_form", name="reviews_add"),
    url(r'^preview$', "preview", name="reviews_preview"),
    url(r'^reedit$', "reedit_or_save", name="reviews_reedit"),
    url(r'^thank-you$', "thank_you", name="reviews_thank_you"),
    url(r'^already-rated$', "already_rated", name="reviews_already_rated"),
)



# django imports
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.core.urlresolvers import reverse
from django.forms import ModelForm
from django.forms.utils import ErrorList
from django.http import HttpResponseRedirect
from django.shortcuts import render_to_response
from django.template import RequestContext
from django.utils.translation import ugettext_lazy as _

# reviews imports
import reviews.signals
from reviews import utils as reviews_utils
from reviews.models import Review
from reviews.settings import SCORE_CHOICES


class ReviewAddForm(ModelForm):
    """Form to add a review.
    """
    class Meta:
        model = Review
        fields = ("user_name", "user_email", "comment", "score")

    def clean(self):
        """
        """
        # For an anonymous user the name is required. Please note that the
        # request has to be passed explicitely to the form object (see add_form)
        msg = _(u"This field is required")

        if self.request.user.is_anonymous():
            if settings.REVIEWS_IS_NAME_REQUIRED:
                if self.cleaned_data.get("user_name", "") == "":
                    self._errors["user_name"] = ErrorList([msg])
            if settings.REVIEWS_IS_EMAIL_REQUIRED:
                if self.cleaned_data.get("user_email", "") == "":
                    self._errors["user_email"] = ErrorList([msg])

        return self.cleaned_data


def add_form(request, content_type_id, content_id, template_name="reviews/review_form.html"):
    """Displays the form to add a review. Dispatches the POST request of the
    form to save or reedit.
    """
    ctype = ContentType.objects.get_for_id(content_type_id)
    object = ctype.get_object_for_this_type(pk=content_id)

    if reviews_utils.has_rated(request, object):
        return HttpResponseRedirect(reverse("reviews_already_rated"))

    scores = []
    for i, score in enumerate(SCORE_CHOICES):
        scores.append({
            "title": score[0],
            "value": score[0],
            "z_index": 10 - i,
            "width": (i + 1) * 25,
        })

    if request.method == "POST":
        form = ReviewAddForm(data=request.POST)
        # "Attach" the request to the form instance in order to get the user
        # out of the request within the clean method of the form (see above).
        form.request = request
        if form.is_valid():
            if settings.REVIEWS_SHOW_PREVIEW:
                return preview(request)
            else:
                return save(request)
    else:
        form = ReviewAddForm()

    return render_to_response(template_name, RequestContext(request, {
        "content_type_id": content_type_id,
        "content_id": content_id,
        "object": object,
        "form": form,
        "scores": scores,
        "show_preview": settings.REVIEWS_SHOW_PREVIEW,
    }))


def reedit(request, template_name="reviews/review_form.html"):

    """Displays a form to edit a review. This is used if a reviewer re-edits
    a review after she has previewed it.
    """
    # get object
    content_type_id = request.POST.get("content_type_id")
    content_id = request.POST.get("content_id")

    ctype = ContentType.objects.get_for_id(content_type_id)
    object = ctype.get_object_for_this_type(pk=content_id)

    if reviews_utils.has_rated(request, object):
        return HttpResponseRedirect(reverse("reviews_already_rated"))

    scores = []
    for i, score in enumerate(SCORE_CHOICES):
        scores.append({
            "title": score[0],
            "value": score[0],
            "current": str(score[0]) == request.POST.get("score"),
            "z_index": 10 - i,
            "width": (i + 1) * 25,
        })

    form = ReviewAddForm(data=request.POST)
    return render_to_response(template_name, RequestContext(request, {
        "content_type_id": content_type_id,
        "content_id": content_id,
        "form": form,
        "scores": scores,
        "object": object,
        "show_preview": settings.REVIEWS_SHOW_PREVIEW,
    }))


def reedit_or_save(request):
    """Edits or saves a review dependend on which button has been pressed.
    """
    if request.POST.get("edit"):
        return reedit(request)
    else:
        return save(request)


def save(request):
    """Saves a review.
    """
    form = ReviewAddForm(data=request.POST)
    form.request = request
    if form.is_valid():
        new_review = form.save(commit=False)
        new_review.content_type_id = request.POST.get("content_type_id")
        new_review.content_id = request.POST.get("content_id")
        new_review.session_id = request.session.session_key
        new_review.ip_address = request.META.get("REMOTE_ADDR")
        if request.user.is_authenticated():
            new_review.user = request.user
        new_review.active = not settings.REVIEWS_IS_MODERATED
        new_review.save()

        # Fire up signal
        reviews.signals.review_added.send(new_review)

        # Save object within session
        ctype = ContentType.objects.get_for_id(new_review.content_type_id)
        object = ctype.get_object_for_this_type(pk=new_review.content_id)
        request.session["last-rated-object"] = object

        return HttpResponseRedirect(reverse("reviews_thank_you"))


def preview(request, template_name="reviews/review_preview.html"):
    """Displays a preview of the review.
    """
    content_type_id = request.POST.get("content_type_id")
    content_id = request.POST.get("content_id")

    ctype = ContentType.objects.get_for_id(content_type_id)
    object = ctype.get_object_for_this_type(pk=content_id)

    if request.user.is_authenticated():
        name = request.user.get_full_name()
        email = request.user.email
    else:
        name = request.POST.get("user_name")
        email = request.POST.get("user_email")

    return render_to_response(template_name, RequestContext(request, {
        "score": float(request.POST.get("score", 0)),
        "object": object,
        "name": name,
        "email": email,
    }))


def thank_you(request, template_name="reviews/thank_you.html"):
    """Displays a thank you page.
    """
    if "last-rated-object" in request.session:
        object = request.session.get("last-rated-object")
        del request.session["last-rated-object"]
    else:
        object = None

    return render_to_response(template_name, RequestContext(request, {
        "object": object,
    }))


def already_rated(request, template_name="reviews/already_rated.html"):
    """Displays a alreday rated page.
    """
    return render_to_response(template_name, RequestContext(request))



# django imports
from django.conf import settings
from django.contrib.auth.models import AnonymousUser
from django.contrib.auth.models import User
from django.contrib.contenttypes.models import ContentType
from django.contrib.flatpages.models import FlatPage
from django.contrib.sessions.backends.file import SessionStore
from django.core.handlers.wsgi import WSGIRequest
from django.core.urlresolvers import reverse
from django.test import TestCase
from django.test import Client

# reviews imports
import reviews.utils
from reviews.models import Review


# Taken from "http://www.djangosnippets.org/snippets/963/"
class RequestFactory(Client):
    """
    """
    def request(self, **request):
        environ = {
            'HTTP_COOKIE': self.cookies,
            'PATH_INFO': '/',
            'QUERY_STRING': '',
            'REQUEST_METHOD': 'GET',
            'SCRIPT_NAME': '',
            'SERVER_NAME': 'testserver',
            'SERVER_PORT': 80,
            'SERVER_PROTOCOL': 'HTTP/1.1',
        }
        environ.update(self.defaults)
        environ.update(request)
        return WSGIRequest(environ)


def create_request(user=None):
    """
    """
    rf = RequestFactory()
    request = rf.get('/')
    request.session = SessionStore()
    if user:
        request.user = user
    else:
        request.user = User(first_name="John Doe", email="john@doe.com")

    return request


class ReviewsModelsTestCase(TestCase):
    """
    """
    def setUp(self):
        """
        """
        # Create a dummy page to test
        self.page = FlatPage.objects.create(url="/test/", title="Test")

    def test_review_defaults(self):
        """
        """
        review = Review.objects.create(content=self.page)

        self.assertEqual(review.content, self.page)
        self.assertEqual(review.score, 3.0)
        self.assertEqual(review.comment, "")
        self.assertEqual(review.active, False)
        self.assertEqual(review.ip_address, None)
        self.assertEqual(review.user, None)
        self.assertEqual(review.session_id, None)
        self.assertEqual(review.user_name, "")
        self.assertEqual(review.user_email, "")

    def test_review_methods(self):
        """
        """
        review = Review.objects.create(
            content=self.page,
            user_name="Jane Doe",
            user_email="jane@doe.com",
        )

        self.assertEqual(review.name, "Jane Doe")
        self.assertEqual(review.email, "jane@doe.com")

        review.user = User(first_name="John", last_name="Doe", email="john@doe.com")

        self.assertEqual(review.name, "John Doe")
        self.assertEqual(review.email, "john@doe.com")

    def test_review_manager(self):
        """
        """
        review_1 = Review.objects.create(content=self.page, creation_date="2009-10-16")
        review_2 = Review.objects.create(content=self.page, creation_date="2009-10-15")

        # all is providing all reviews
        result = Review.objects.all()
        self.assertEqual(len(result), 2)

        # active is providing just the active ones
        result = Review.objects.active()
        self.assertEqual(len(result), 0)

        review_1.active = True
        review_1.save()

        # active is providing just the active ones
        result = Review.objects.active()
        self.assertEqual(len(result), 1)

        review_2.active = True
        review_2.save()

        # active is providing just the active ones
        result = Review.objects.active()
        self.assertEqual(len(result), 2)

    def test_review_ordering(self):
        """
        """
        Review.objects.create(content=self.page, id=1)
        Review.objects.create(content=self.page, id=2)
        Review.objects.create(content=self.page, id=3)

        # Last in first out
        result = [r.id for r in Review.objects.all()]
        self.assertEqual(result, [3, 2, 1])


class ReviewsViewsTestCase(TestCase):
    """
    """
    def setUp(self):
        """
        """
        # Create a dummy page to test
        self.page = FlatPage.objects.create(url="/test/", title="Test")

    def test_add_form(self):
        """
        """
        ctype = ContentType.objects.get_for_model(self.page)
        url = reverse("reviews_add", kwargs={"content_type_id": ctype.id, "content_id": self.page.id})

        # The result has to ``Preview`` within it
        result = self.client.get(url)
        self.assertContains(result, "Preview", status_code=200)

        # switching ot no preview
        settings.REVIEWS_SHOW_PREVIEW = False

        # The result must not have ``Preview`` within it
        result = self.client.get(url)
        self.assertNotContains(result, "Preview", status_code=200)


class PortletsUtilsTestCase(TestCase):
    """
    """
    def setUp(self):
        """
        """
        # Create a dummy page to test
        self.page = FlatPage.objects.create(url="/test/", title="Test")

    def test_get_average_for_instance(self):
        """
        """
        # Add a review to the page
        self.review = Review.objects.create(content=self.page, score=4.0)

        # By default the review is not active so there is no average rating
        average = reviews.utils.get_average_for_instance(self.page)
        self.assertEqual(average, (None, 0))

        # Make the review active
        self.review.active = True
        self.review.save()

        average = reviews.utils.get_average_for_instance(self.page)
        self.assertEqual(average, (4.0, 1))

        # Now we add another one
        Review.objects.create(content=self.page, score=2.0, active=True)

        average = reviews.utils.get_average_for_instance(self.page)
        self.assertEqual(average, (3.0, 2))

    def test_get_reviews_for_instance(self):
        """
        """
        # Add a review to the page
        self.review = Review.objects.create(content=self.page, score=4.0)

        # By default the review is not active so there is no average rating
        result = reviews.utils.get_reviews_for_instance(self.page)
        self.assertEqual(len(result), 0)

        # Make the review active
        self.review.active = True
        self.review.save()

        result = reviews.utils.get_reviews_for_instance(self.page)
        self.assertEqual(len(result), 1)

        # Now we add another one
        Review.objects.create(content=self.page, score=2.0, active=True)

        result = reviews.utils.get_reviews_for_instance(self.page)
        self.assertEqual(len(result), 2)

    def test_get_best_rated_for_model(self):
        """
        """
        # Create some dummy pages
        self.page_1 = FlatPage.objects.create(url="/test-1/", title="Test 1")

        # There are no reviews for pages, so it should return None
        result = reviews.utils.get_best_rated_for_model(self.page)
        self.assertEqual(result, None)

        # Start to add reviews
        Review.objects.create(content=self.page_1, score=4.0, active=True)

        self.page_2 = FlatPage.objects.create(url="/test-2/", title="Test 2")
        Review.objects.create(content=self.page_2, score=3.0, active=True)

        # At first page 1 is best
        result = reviews.utils.get_best_rated_for_model(self.page_1)
        self.assertEqual(result[0], self.page_1)

        # Adding one more review to page 2, but page 1 is still the best
        Review.objects.create(content=self.page_2, score=4.0, active=True)
        result = reviews.utils.get_best_rated_for_model(self.page_1)
        self.assertEqual(result[0], self.page_1)

        # Adding one more review to page 2. Now page 2 is the best rated
        Review.objects.create(content=self.page_2, score=6.0, active=True)
        result = reviews.utils.get_best_rated_for_model(self.page_1)
        self.assertEqual(result[0], self.page_2)

    def test_get_best_rated(self):
        """
        """
        # Create some dummy pages
        self.page_1 = FlatPage.objects.create(url="/test-1/", title="Test 1")

        # There are no reviews for pages, so it should return None
        result = reviews.utils.get_best_rated()
        self.assertEqual(result, None)

        # Start to add reviews
        Review.objects.create(content=self.page_1, score=4.0, active=True)

        self.page_2 = FlatPage.objects.create(url="/test-2/", title="Test 2")
        Review.objects.create(content=self.page_2, score=3.0, active=True)

        # At first page 1 is best
        result = reviews.utils.get_best_rated()
        self.assertEqual(result[0], self.page_1)

        # Adding one more review to page 2, but page 1 is still the best
        Review.objects.create(content=self.page_2, score=4.0, active=True)
        result = reviews.utils.get_best_rated()
        self.assertEqual(result[0], self.page_1)

        # Adding one more review to page 2. Now page 2 is the best rated
        Review.objects.create(content=self.page_2, score=6.0, active=True)
        result = reviews.utils.get_best_rated()
        self.assertEqual(result[0], self.page_2)

    def test_has_rated(self):
        """
        """
        request = create_request()
        request.user = AnonymousUser()

        # Create some dummy pages
        self.page_1 = FlatPage.objects.create(url="/test-1/", title="Test 1")
        self.page_2 = FlatPage.objects.create(url="/test-2/", title="Test 2")

        # At first nobody has rated
        self.assertEqual(reviews.utils.has_rated(request, self.page_1), False)
        self.assertEqual(reviews.utils.has_rated(request, self.page_2), False)

        # Rate for page 1
        Review.objects.create(content=self.page_1, score=6.0, active=True, session_id=request.session.session_key)
        self.assertEqual(reviews.utils.has_rated(request, self.page_1), True)
        self.assertEqual(reviews.utils.has_rated(request, self.page_2), False)

        # Rate for page 2
        Review.objects.create(content=self.page_2, score=6.0, active=True, session_id=request.session.session_key)
        self.assertEqual(reviews.utils.has_rated(request, self.page_1), True)
        self.assertEqual(reviews.utils.has_rated(request, self.page_2), True)
